//  Copyright (c) 1997-2000 by Mark F. Adams 
//  $Id: geo_grid.C,v 1.13 2006/05/19 20:48:09 adams Exp $
//  Author:             Mark F. Adams
//  Version:            2.0
//  Filename:           geo_grid.C
//
#include "prom_grid.hh"

//! fine PromGeoGrid constructor (const dof per node)
#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::PromGeoGrid" 
PromGeoGrid::PromGeoGrid( const PromOptions &opts, PromPerfMonitor &perf, 
			  PromParentBank * const bank, const int nloc, 
			  const PromTable<int> &gid_ghostLid,
			  const PromList<int*> *elemList, 
			  const PromComm &wcomm, int bsize ) :
  PromGrid(opts,perf,bank,nloc,wcomm,bsize),   elemList_(50)
{
  int ierr = InitNodes( nloc+gid_ghostLid.getCount() ); assert(proc_geq_!=NULL);
  if( ierr == 0 ) ierr = FineInit( NULL, bsize, gid_ghostLid, elemList );
  if( ierr == 0 ) ierr = BuildElemList( gid_ghostLid, elemList );
  if( ierr == 0 ) ierr = InitEnvNodes();
  if( ierr == 0 ) ierr = MakeElemAdjacs();
  
  // output THIS grid
  if( options_.files_ && isActive() ){ 
    assert( !elemList_.isEmpty() );
    WriteFEAP( 10000*getLevel()+comm_.mype(), gComm_, &elemList_ );
  }
}

//! normal coarse grid constructor (const dof per node)
PromGeoGrid::PromGeoGrid( const PromOptions &opts,
			  PromPerfMonitor &perf, 
			  PromGeoGrid *lastgrid, const int nloc, 
			  const int nnodes, 
			  const PromComm &wcomm, int bsize ) :
  PromGrid(opts, perf, lastgrid->parBank_, nloc, wcomm, bsize),
  elemList_(50)
{
  int ierr = InitNodes( nnodes );
  if( ierr == 0 ) ierr = InitNormal( lastgrid );
  if( ierr == 0 ) ierr = InitEnvNodes();
}

//! dummy coarse grid constructor (const dof per node)
PromGeoGrid::PromGeoGrid( const PromOptions &opts,
			  PromPerfMonitor &perf, PromGeoGrid *lastgrid,
			  const int fact, const PromComm &wcomm, int bsize ) :
  PromGrid( opts, perf, lastgrid->parBank_, 0, fact, wcomm, bsize ),
  elemList_(1)
{
  int ierr = InitNodes( 0 );
  if( ierr == 0 ) ierr = InitDummy( lastgrid );
  if( ierr == 0 ) ierr = InitEnvNodes();
}

//! Reduced coarse grid constructor (splitter) (const dof per node)
PromGeoGrid::PromGeoGrid( const PromOptions &opts,
			  PromPerfMonitor &perf, 
			  PromGeoGrid *lastgrid, const int factor, 
			  const int buffsz, int *buffer, const int rtag, 
			  PromTable<int> &gid1_lid,
			  PromList<PromMPINode*> &ghostlist, 
			  const int partloc, const int nnodes, 
			  const PromComm &wcomm, int bsize ) : 
  PromGrid( opts, perf, lastgrid->parBank_, partloc, factor, wcomm, bsize ),
  elemList_(50)
{
  int ierr = InitNodes( nnodes );
  assert( lastgrid && lastgrid->isActive() );
  
  if(ierr == 0) ierr = InitReduced( lastgrid, factor, buffsz, buffer, rtag,
				    gid1_lid, ghostlist );
  
  if(ierr == 0 ) ierr = InitEnvNodes();
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::InitEnvNodes" 
/* PromGeoGrid::InitEnvNodes ******************************************
 *
 *  - private constructor
 */  
int PromGeoGrid::InitEnvNodes()
{ 
  if( nNodes_ ) {
    for( int xx = 0 ; xx < PROM_NUMENV ; xx++ ){
      PromNode *curr = env_[xx] = newGeoNode();
      // modify envelope elements
      curr->setEnvelope();
      curr->setState( PROM_NOTUSED ); // let envs be NOTUSED
      for( int jj=0 ; jj < PROM_MAXNDF ; jj++ ) curr->setFixed(jj,FALSE);
    }
  }
  else for( int xx = 0 ; xx < PROM_NUMENV ; xx++ ) env_[xx] = NULL;

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::SetCoords" 
/* PromGeoGrid::SetCoords ******************************************
 *
 *  - fine grid private constructor
 */  
int PromGeoGrid::SetCoords(const double crds[], const int ndm,
			   const double * const deltas /*=NULL*/) const
{ 
  if( crds != NULL ) { // fine grid
    for( int ii = 0, kk = 0 ; ii < nNodes_ ; ii++ ) {
      PromGeoNode * curr = dynamic_cast<PromGeoNode*>(nodes_[ii]);
      for( int jj = 0 ; jj < ndm ; jj++, kk++ ) {
	double temp = crds[kk] + (deltas ? deltas[kk] : 0. );
	curr->symbGeoData_->coord[jj] = temp;
      }
    }
  }
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::SetVects" 
/* PromGeoGrid::SetVects ******************************************
 *
 *  - fine grid private constructor
 */  
int PromGeoGrid::SetVects(const double crds[], const int ndm )
{ 
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::~PromGeoGrid" 
/* PromGeoGrid::~PromGeoGrid ******************************************
 *
 */  
PromGeoGrid::~PromGeoGrid() 
{
  int i;
  for(i=0;i<PROM_NUMENV;i++) if(env_[i] != NULL){Kill(env_[i]); env_[i]=NULL;}

  // remove all traces of elements
  DeleteElemList();
  DeleteTets();
  DeleteFacets();
}

int PromGeoGrid::CreateCoarseGridSpace( PromGrid *lastgrid )
{
  return CreateGeometricCoarseSpace( dynamic_cast<PromGeoGrid*>(lastgrid) );
}

int PromGeoGrid::FineGridInitWithA() 
{
  bool flag; int ierr;
 
  ierr = options_.HasOptionName("-geomg_use_geo_hueristics",&flag);
  CHKERRQ(ierr);
  if( flag ) { 
    ierr = SetNodeTypes( comm_.getNewTag() );  CHKERRQ(ierr);
  }
  return 0;
}

PromNode *PromGeoGrid::newNode( const int lid )
{
  PromGeoNode *curr = newGeoNode();
  curr->index_ = lid;
  
  return curr;
}

PromMGType PromGeoGrid::getMGType()
{ 
  return PROM_GEOMG; 
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::findNodeElems"
/* PromGeoGrid::findNodeElems **********************************************
 *
 *  - goes through list, looks at parents elements, for LOCAL nodes
 *
 *  INPUT:
 *     - lost: number of unfound nodes (out)
 *
 *   SIDE resets:
 *     - EFFECTSg all nodes shape values  
 *     - set nodes 'elem'
 * 
 *  RETURN:
 *     - num not found
 */
int PromGeoGrid::findNodeElems ( int &lost ) const
{ 
  PromGeoNode  *fine, *curr;
  PromDTet     *elem; 
  int           ierr, mype = comm_.mype(); 
  gfloat        fout, fbest;
  PromDTet     *temp, *best;
  MPI_Comm      comm = (options_.verbose_>1) ? PETSC_COMM_SELF : MPI_COMM_NULL;
  
  // find nodes in elements, keep unfound nodes in list
  lost = 0; 
  for ( int xx = 0 ; xx < nLocalNd_ ; xx++ ){
    fine = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
    if( fine->getState() == PROM_NOTUSED )  continue; 
    assert( !fine->symbGeoData_->parElem ); 
    best = NULL; fbest = 1.e10;
    
    // find it's element
    curr = dynamic_cast<PromGeoNode*>(fine->getParent()); 
    if( curr && !curr->symbData_->elemList.isEmpty() ) {
      elem=fine->setElemWithList( &curr->symbData_->elemList,  parBank_, mype,
				  5, 1.e-4, TRUE, TRUE );
      if( !elem ) {
	PetscPrintf(comm,"\t[%d] %d not found -- try harder\n",
		    mype,xx);
	elem = fine->setElemWithList( &curr->symbData_->elemList, parBank_, 
				      mype, 12, 1.e-4, TRUE, TRUE );
      }
      else assert(fine->symbGeoData_->parElem!=NULL);
    }
    if( !fine->symbGeoData_->parElem ){
      PetscPrintf(comm,"\t[%d] try real harder\n",mype,xx);
      Position posx = elemList_.getHeadPosition(); 
      while ( posx && !fine->symbGeoData_->parElem ) {
	elem = dynamic_cast<PromDTet*>(elemList_.getNext(posx));
	if( !fine->isChildListEmpty() ) {
	  PetscPrintf(comm,"[%d]ERROR coarse=%d/%d, fine=%d\n",
		      mype,xx,nLocalNd_,fine->index_(ierr));
	  assert( fine->isChildListEmpty() );
	}
	if( !elem->isDead() ) {
	  fine->setElem( elem,parBank_,mype,1.e-4,TRUE, 1, NULL, TRUE, &temp,
			 &fout );
	  if( temp && (!best || fout < fbest) ) { fbest = fout; best = temp; }
	}
      }
      if(!best && !fine->symbGeoData_->parElem)SETERRQ(1,"elem !found!!!");
      if( !fine->symbGeoData_->parElem ) {
	assert( fine->isChildListEmpty() );
	fine->setElem( best, parBank_, mype,1.e10, TRUE, 0, NULL, TRUE );
	PetscPrintf(comm,"\t\t[%d] %d very hard find\n",
		    mype,xx);
	assert( fine->symbGeoData_->parElem != NULL );
      }
    }
    assert( fine->symbGeoData_->parElem && fine->nParents() > 0 );
  } // all fine nodes

  return 0;
}
        
#undef __FUNCT__    
#define __FUNCT__ "PromGrid::WriteFEAP" 
////////////////////////////////////////////////////////////////////////
//
//  Write FEAP functiuons
//
/* PromGrid::WriteFEAP ****************************************************
 *
 *  - write FEAP text input file - (tris and tets)
 * 
 *  INPUT:
 *    - lev: level (appended to file name
 *    - objects elem list and node list
 *
 *   SIDE EFFECTS:
 *     - writes a file  
 *
 */
int PromGrid::WriteFEAP( int lev, MPI_Comm comm, 
			    PromList<PromPoly*>* elist,
			    PromList<PromFacet*>* flist/*=NULL*/)const 
{ 
  FILE      *file, *nodefile, *elemfile; 
  char       path[256];
  int        xx,mype=comm_.mype();
  
  //sprintf ( path, "/users/madams/feap/%s%d",name, level);
  sprintf ( path, "igrid_%d", lev );
  if( !(file = fopen(path, "w")) ) return 1;  
  sprintf ( path, "coor_%d", lev );
  if( !(nodefile = fopen(path, "w")) ) return 1;
  sprintf ( path, "elem_%d", lev );
  if( !(elemfile = fopen(path, "w")) ) return 1;
  
  fprintf(file,"feap\n");
  fprintf(file, "%d %d 0 %d %d %d\n\nnopa\n\n",0,0,3,3,8);
 
  // COOR: write nodes
  int m1 = (proc_gnode_==NULL) ? 1 : proc_gnode_[mype] + 1;
  fprintf(file,"\n\ncoor\n");
  PromCoord crd;
  PromGeoNode *curr; 
  for ( xx = 0 ; xx < nNodes_ ; xx++ ){
    if( xx < nNodes_ ) curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);
    crd = curr->symbGeoData_->coord;
    fprintf(file,"%d 0 %le %le %le\n", xx+1, crd[0], crd[1], crd[2] );
    if(xx<nNodes_)
      fprintf(nodefile,"%d 0 %le %le %le\n",m1+xx,crd[0],crd[1],crd[2]);
  }
  assert ( !(elist && flist) );
  if ( !flist ) {
    WriteElems( file, elemfile, elist, comm );
  }
  //else   WriteFacets( file, flist );
 
  // BOUN
  fprintf(file,"\n\nboundry: x  y  z\n");
  for ( xx = 0 ; xx < nNodes_ ; xx++ ){
    curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);
    if ( curr->isFixed() ) {
      fprintf(file,"%d 0", xx+1);
      for ( int j = 0 ; j < 3 ; j++ )
        fprintf(file," %d",curr->isFixed(j));
      fprintf(file,"\n");
    }
  }
    
  WriteMATE( file );
  
  fclose(file);
  fclose(nodefile);
  fclose(elemfile);

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::WriteElems" 
/* PromGrid::WriteElems **************************************************
 *
 *  - 
 *
 *  INPUT:
 *    - file: output file
 *    - efile: global outputfile
 *    - elems:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::WriteElems ( FILE *file, FILE *efile, 
			   const PromList<PromPoly*> * const elems,
			   MPI_Comm comm ) const
{
  int nelems, i, ierr;
  Position pos;
  const PromPoly *s;

  // get my first element number
  pos = elems->getHeadPosition(); nelems = 0;
  while( pos ) {
    s = elems->getNext(pos);
    if( !s->isDead() ) nelems++;
  } 
  int *proc_elem = NULL;
  ierr = PromContext::ScanArrMPI( nelems, proc_elem, comm ); CHKERRQ(ierr);
  int m0 = proc_elem ? proc_elem[comm_.mype()] + 1 : 0;
  if(proc_elem){ PetscFree(proc_elem); }
  proc_elem = NULL;
  
  // ELEM
  fprintf( file, "\n\nelem ! %d\n", nelems );
  pos = elems->getHeadPosition(); i = 1;   assert(pos!=NULL);
  while ( pos ) {
    s = elems->getNext( pos ); 
    //if( s->NumEnv() ) continue;
    fprintf(file,"%d 0 %d",i++,(s->getMaterialID()<1) ? 2 :s->getMaterialID());
    s->WriteFEAP( file );
    fprintf(efile,"%d 0 %d",m0++,isBase() ? s->getMaterialID():comm_.mype()+1);
    s->WriteFEAP( efile, proc_gnode_ );
  }
  
  return 0;
} 

#undef __FUNCT__    
#define __FUNCT__ "PromPoly::WriteFEAP" 
/* PromPoly::WriteFEAP() ***********************************************
 *
 */
int PromPoly::WriteFEAP( FILE *file, const int pg[] /*=NULL*/) const
{
  int mype; MPI_Comm_rank( PETSC_COMM_WORLD, &mype ); 
  int jj, en[8], m0 = pg ? pg[mype] + 1 : 1; 
  assert(numNodes()<=8);
   
  for( jj = 0 ; jj < numNodes() ; jj++ ) {
    if( nodes_[jj]->isGhost() && pg ) 
      en[jj] = nodes_[jj]->symbData_->ghostID.local() + 
	pg[nodes_[jj]->symbData_->ghostID.proc()] + 1;
    else en[jj] =  nodes_[jj]->index_() + m0;
  }
  
  if( numNodes() == 8 ) {
    for( jj = 0 ; jj < 8 ; jj++ ) {
      fprintf( file," %d", en[jj] );
    }
    fprintf( file, "\n" );
  }
  else if( numNodes() == 4 ) { 
    fprintf( file," %d %d %d %d\n", en[0], en[1], en[2], en[3] );
  }
  else {
    for( jj = 0 ; jj < numNodes() ; jj++ ) {
      fprintf(file," %d", en[jj] );
    }
    fprintf( file, "\n" );
  }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::WriteFacets" 
/* PromGrid::WriteFacets **************************************************
 *
 *  - 
 *
 *  INPUT:
 *    - file: output file
 *
 *  SIDE EFFECTS:
 *    - writes to file 
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::WriteFacets ( FILE *file, PromList<PromFacet*> *facets ) const
{
  // ELEM
  fprintf(file,"\n\nelements");
  int i = 0;
  Position pos = facets->getHeadPosition(); 
  while ( pos ) {
    const PromFacet *s = facets->getNext( pos );
    
    fprintf(file,"\n%d 0 %d",++i,(s->getMaterialID()<1)? 2:s->getMaterialID());
    const int numn = s->nodes_[3] ? 4 : 3;
    for ( int jj = 0 ; jj < numn ; jj++ ) {
      fprintf(file," %d", s->nodes_[jj]->index_()+1);
    }
  }
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::WriteFEAP" 
/* PromGrid::WriteFEAP ****************************************************
 *
 *  - write FEAP text input file - (tris and tets)
 * 
 *  INPUT:
 *    - lev: level (appended to file name
 *    - elems: list of elems
 *
 *   SIDE EFFECTS:
 *     - writes a file
 *
 *   RETURN:
 *     - PETSc error code
 */
int PromGrid::WriteFEAP( int lev, PromTable<PromPoly*> *elems )const
{
  int        i;
  TablePos   tpos;
  PromGeoNode *curr;
  FILE      *file;
  PromPoly  *s;
  char       path[256];
  
  //sprintf ( path, "/users/madams/feap/%s%d",name, level);
  sprintf( path, "igrid%d", lev );
  if( !(file = fopen(path, "w")) ) SETERRQ(1,"fopen error");
  
  fprintf(file,"feap\n");
  const int nen = 8,ndf=3; 
  fprintf(file,"0 0 0 %d %d %d\n",ndf,3,nen);
  
  // COOR: write nodes
  fprintf(file,"\n\ncoor\n");
  for ( i = 0 ; i < nNodes_ ; i++ ) {
    curr = dynamic_cast<PromGeoNode*>(nodes_[i]);
    fprintf(file,"%d 0 %le %le %le\n",i+1,
            curr->symbGeoData_->coord[0],curr->symbGeoData_->coord[1],
	    curr->symbGeoData_->coord[2]);
  }
  
  // ELEM
  fprintf(file,"\n\nelements\n");
  i = 0;
  tpos = elems->getHeadPosition();
  while (tpos) {
    s = elems->getNext(tpos);
    fprintf(file,"%d 0 %d",++i,(s->getMaterialID()<1) ? 2:s->getMaterialID());
    s->WriteFEAP( file );
  }
  
  WriteMATE( file );
  
  fclose(file);
  
  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGrid::WriteMATE" 
/* PromGrid::WriteMATE **************************************************
 *
 *  - write end of feap file.
 *
 *  INPUT:
 *    - file: output file
 *
 *  SIDE EFFECTS:
 *    - sets exterior nodes
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGrid::WriteMATE ( FILE *file )const
{
  // MATE
  fprintf(file,"\n\nincl,mate\n\nend\n\nmacr\nplot,colo\nplot,defa,off\nend\n\ninte\n\nstop\n");
  return 0;
}

////////////////////////////////////////////////////////////////////////
// node typing routines
//
static shpfloat COSTOL = .5;
#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::SetNodeTypes" 
/* PromGeoGrid::SetNodeTypes ******************************************
 *
 *  - set node topotype
 * 
 *  INPUT:  
 *  SIDE EFFECTS:
 *    - sets node types, facet and surface
 *
 *  RETURN:
 *    - PETSc error code
 */  
int PromGeoGrid::SetNodeTypes( const int tag /*-1*/ )
{ 
  int     ierr, xx,jj,tt; 
  Position    posx;
  PromGeoNode  *curr;
  PromList<PromFacet*> *flists = NULL, facets(100);
  MPI_Comm comm = options_.verbose_ > 1 ? 
    ((prevgrid_==NULL) ? gComm_ : prevgrid_->gComm_) : MPI_COMM_NULL;
  if( prevgrid_ != NULL && !prevgrid_->isActive() ) return 0;

  if( options_.verbose_ > 1 )
    PetscPrintf(comm,"[%d]%s Set node types\n",comm_.mype(),__FUNCT__);
  { // check for user provided tolerence on angles  
    bool flag; double dd;
    ierr = options_.GetOptionDouble("-geomg_cos_tol",&dd,&flag);CHKERRQ(ierr);
    if( flag ){
      COSTOL = dd;
      if( options_.verbose_ > 1 )
	PetscPrintf(comm,"[%d]%s set COSTOL=%e\n",comm_.mype(),__FUNCT__,dd);
    }
  }

  xx = nNodes_;
  if( xx > 0 && (ierr=MakeFacets(&facets)) == 0 && !facets.isEmpty() ) {
    // put faces in nodes 'flists'
    flists = new PromList<PromFacet*>[xx];   CHKNEW(flists,1);
    posx = facets.getHeadPosition();
    while( posx ) {
      PromFacet *facet = facets.getNext(posx);
      const int N1 = facet->nodes_[3] ? 4 : 3;
      for( jj = 0 ; jj < N1 && facet->nodes_[jj]; jj++ ) {
	curr = dynamic_cast<PromGeoNode*>(facet->nodes_[jj]);
	assert(curr!=NULL);
	tt = curr->index_(ierr); 	assert(tt >= 0 && tt < xx);
	flists[tt].addTail( facet );
      }
      assert(jj>2);
    }
  }
  
  // make types
  if( tag < 0 ) { // only globalize first 
    // set colors (facetID) on facetList, empty and add to seedFacets
    ierr = SetFaceIDs( facets, flists );                        CHKERRQ(ierr);
    // set types
    ierr = SetTopoTypes( flists );                              CHKERRQ(ierr);
  }
  else if( isActive() ) {
    PromTable<int> lwid_hgid_alias; 
    PromList<int> higherProcs, lowerProcs; PromList<PromFacet*> seedFacets;
    // make proc table
    for( int xx = nLocalNd_ ; xx < nNodes_ ; xx++ ) {
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
      int proc = curr->symbData_->ghostID.proc();
      int add = lwid_hgid_alias.add(proc+1);
      if( add ){
	if( proc > comm_.mype()) higherProcs.addTail( proc+1 );
	else lowerProcs.addTail(proc+1);
      }
    }
    lwid_hgid_alias.removeAll();
    // wait for others
    if( higherProcs.isEmpty() && options_.verbose_ > 1 ) {
      PetscPrintf( PETSC_COMM_SELF,"[%d]%s root proc\n",comm_.mype(),__FUNCT__);
    }
    if( options_.verbose_ > 2 ){ 
      jj = higherProcs.getCount() + lowerProcs.getCount();
      ierr = MPI_Allreduce(&jj,&tt,1,MPI_INT, MPI_MAX, gComm_ ); CHKERRQ(ierr);
      PetscPrintf(gComm_,"[%d]%s %d MAX proc adj\n",comm_.mype(),__FUNCT__,tt);
      jj = facets.isEmpty();
      ierr = MPI_Allreduce(&jj,&tt,1, MPI_INT, MPI_SUM, gComm_); CHKERRQ(ierr);
      PetscPrintf(gComm_,"[%d]%s %d interior dom\n",comm_.mype(),__FUNCT__,tt);
    }
    ierr = CollectSeeds( tag, higherProcs, facets, seedFacets ); CHKERRQ(ierr);
    
    if( flists ){
      // set colors (facetID) on facetList - add to seedFacets
      ierr = SetFaceIDs( facets, flists, &tt, &seedFacets, &lwid_hgid_alias ); 
      CHKERRQ(ierr);
    }
    else tt = 0;
    
    // make and send message to lower
    ierr = SendSeeds( tag, lowerProcs, facets );    CHKERRQ(ierr);    

    MPI_Reduce( &tt, &jj, 1, MPI_INT, MPI_SUM, 0, gComm_);
    if(options_.verbose_>3)
      PetscPrintf(gComm_,"[%d]%s Made %d PREGLOBAL faux faces\n",
		  comm_.mype(), __FUNCT__, jj );
    
    // globalize facids
    ierr = GlobalizeFacets( facets, lwid_hgid_alias ); CHKERRQ(ierr);
    
    // reset norms
    Position pos = facets.getHeadPosition(); 
    while( pos ){
      PromFacet *facet = facets.getNext(pos);
      assert(facet->getID() > 0);
      PromContext::cross((dynamic_cast<PromGeoNode*>(facet->nodes_[0]))->symbGeoData_->coord,
			 (dynamic_cast<PromGeoNode*>(facet->nodes_[1]))->symbGeoData_->coord,
			 (dynamic_cast<PromGeoNode*>(facet->nodes_[2]))->symbGeoData_->coord,
			 &facet->norm_ ); 
      facet->norm_ = facet->norm_ / facet->norm_.length();
    }
    
    // make types
    ierr = SetTopoTypes( flists );
    CHKERRQ(ierr);
  }

  // clean up faces
  DeleteFacets();
  if( flists ){ delete [] flists; }

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::MakeFacets" 
/* PromGeoGrid::MakeFacets ******************************************
 *
 *  - set make facet list with null (and diff. mat.) 
 *
 *  INPUT: 
 *    - facets: list to put new facets
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::MakeFacets( PromList<PromFacet*> *facets )
{
  Position   posx;
  PromPoly   *elem;
  PromFacet     *facet;
  PromGeoNode *n1,*n2,*n3,*n4;
  double len;

  // make face list
  posx = elemList_.getHeadPosition(); 
  while ( posx ) {
    elem = elemList_.getNext(posx); 
    if ( elem->getType() == PROM_PLANE ) { 
      n1 = dynamic_cast<PromGeoNode*>(elem->nodes_[0]); n1->setExterior();   
      n2 = dynamic_cast<PromGeoNode*>(elem->nodes_[1]); n2->setExterior();   
      n3 = dynamic_cast<PromGeoNode*>(elem->nodes_[2]); n3->setExterior();  
      n4 = dynamic_cast<PromGeoNode*>(elem->nodes_[3]);
      if(n4) n4->setExterior();  
      facet = newFacet( n3, n2, n1, n4 );
      len = facet->norm_.length();
      if( len == 0.0 ) {
	n1->Archive( stderr, PROM_PRINT );
	n2->Archive( stderr, PROM_PRINT );
	n3->Archive( stderr, PROM_PRINT );
	if(n4) n4->Archive( stderr, PROM_PRINT );
	SETERRQ(1,"facet on PLANE is degenerate");
      }
      facet->norm_ = facet->norm_ / len ; // normailze
      facets->addTail( facet );
      facet->setMaterialID( elem->getMaterialID() ); 
    }
    else if ( elem->getType() == PROM_N_POLY ){
      // ignore others
      for ( int ii = 0 ; ii <  elem->numNodes() ; ii++ ) {
	elem->nodes_[ii]->setExterior();   
      }
    }              // hack to ignore "tets" (could be membranes)
    else if( elem->getType() == PROM_SOLID && elem->numNodes() == 8 ){
      const int numadj = elem->numNeigh(); 
      const int *elID = elem->getIDArr();
      // look for a face w/o a nieghbor (with its material) - make a face
      for ( int ii = 0 ; ii < numadj ; ii++ ) {
        // add one face per region boundry
        if(!elem->neigh_[ii] || 
	   elem->neigh_[ii]->getMaterialID()>elem->getMaterialID()){
          // face nodes - reverse order - poining out
          n1 = dynamic_cast<PromGeoNode*>(elem->nodes_[elID[4*ii+2]]); 
	  n2 = dynamic_cast<PromGeoNode*>(elem->nodes_[elID[4*ii+1]]); 
          n3 = dynamic_cast<PromGeoNode*>(elem->nodes_[elID[4*ii]  ]);  
	  if ( elID[4*ii+3] < 0 ) n4 = NULL;
	  else n4 = dynamic_cast<PromGeoNode*>(elem->nodes_[elID[4*ii+3]]);
	  // set exterior nodes here
	  if( !elem->neigh_[ii] ) {
	    n1->setExterior();   n2->setExterior();   
	    n3->setExterior();   if( n4 ) n4->setExterior();   
	  }
          if( !n1->isGhost() || !n2->isGhost() || !n3->isGhost() || 
	      (n4 && !n4->isGhost()) ) {
	    facet = newFacet( n1, n2, n3, n4 );
	    len = facet->norm_.length();
	    if( len == 0.0 ) {
	      n1->Archive( stderr, PROM_PRINT );
	      n2->Archive( stderr, PROM_PRINT );
	      n3->Archive( stderr, PROM_PRINT );
	      if(n4) n4->Archive( stderr, PROM_PRINT );
	      SETERRQ4(1,"facet on PLANE is degenerate isg=%d,%d,%d,%d",
		       n1->isGhost(),n2->isGhost(),
		       n3->isGhost(), n4 ? n4->isGhost() : -1);
	    }
	    facet->norm_ = facet->norm_ / len; // normailze
	    assert( !facet->isMaster() );
	    facets->addTail( facet );
	    facet->setMaterialID( elem->getMaterialID() );
	    if ( elem->neigh_[ii] )facet->setExterior(FALSE);//interior boundry
	    else facet->setExterior();
	  }
	}
      }
    }
  }  // elems

  return 0;
}
 
#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::CollectSeeds" 
/* PromGeoGrid::CollectSeeds ******************************************
 *
 *  - collect seeds from higher procs
 * 
 *  INPUT: 
 *    - tag: MPI_ANY_SOURCE
 *    - higherProcs: 
 *    - facets:
 *    - seedFacets: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::CollectSeeds( const int tag, PromList<int> &higherProcs, 
			       PromList<PromFacet*> &facets, 
			       PromList<PromFacet*> &seedFacets)const
{
  int        ierr,bsz, *pb; 
  MPI_Comm comm = (prevgrid_ == NULL) ? gComm_ : prevgrid_->gComm_;
  const int comfact = (prevgrid_==NULL) ? factor_ : prevgrid_->factor_; 
  MPI_Status status; 
  PromCoord      crd;
  int __rbuf[PROM_BSZ+1]; 
  int *rbuf = ((size_t)__rbuf%sizeof(double)==0) ? __rbuf: __rbuf+1; 

  while( !higherProcs.isEmpty() ){  
    // recv(tag) < norm, matid, faceid, *<n1, n2, n3> >
    ierr = MPI_Recv(rbuf, PROM_BSZ, MPI_INT, MPI_ANY_SOURCE,tag,comm,&status);
    CHKERRQ(ierr);
    const int proc = status.MPI_SOURCE*comfact;
    MPI_Get_count( &status, MPI_INT, &bsz ); 
    if( bsz == 0 ) {
      Position pos2 = higherProcs.find(proc+1); assert(pos2!=NULL);
      higherProcs.removeAt(pos2);
      continue;
    }
    // recieve
    pb = rbuf; assert( (size_t)pb%sizeof(gfloat) == 0 );
    // norm
    crd.Read( &pb ); 
    const int mid = *pb++, fid = *pb++;               assert(fid>0);
    // find facid and add to seeds,...
    while( pb - rbuf < bsz ){
      int n1 = *pb++, n2 = *pb++, n3 = *pb++;
      // find facet
      PromFacet *facet = NULL;
      Position pos = facets.getHeadPosition(); // linear search !!!
      while( pos && !facet ) {
	facet = facets.getNext(pos);
	if( facet->nodes_[0]->index_.global() == n1 &&
	    facet->nodes_[1]->index_.global() == n2 &&
	    facet->nodes_[2]->index_.global() == n3 ) break;
	else facet = NULL;
      }
      if( !facet ) SETERRQ(1,"!facet");
      
      // add it
      assert( !facet->isMarked() );
      if( !facet->isMaster() ) {
	facet->setMaster(); 
	seedFacets.addHead( facet ); assert( facet->getID() <= 0 );
	facet->norm_ = crd;
	assert( facet->getMaterialID() == mid );
	facet->setFaceID( fid ); assert(fid>0);
      }
      else assert(facet->getID() == fid);
    } // each facet
    assert(pb-rbuf == bsz);
  }
  
  return 0;
}

static int PROM_MAX_FACE_ID = 0;
#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::SetFaceIDs" 
/* PromGeoGrid::SetFaceIDs *****************************************
 *  - set node types
 *
 *  INPUT:
 *    - facets: list of all facets
 *    - flists:  
 *    - nfaces: (out)
 *    - seedFacets: 
 *    - lwid_hgid_alias: (out)
 *
 *  SIDE EFFECTS:
 *    - empty 'facets'
 *    - sets face IDs of facets in 'faces'
 *
 *  RETURN:
 *    - 
 */ 
int PromGeoGrid::SetFaceIDs( PromList<PromFacet*> &facets,
			     PromList<PromFacet*> flists[], 
			     int *nfaces/*=NULL*/, 
			     PromList<PromFacet*> *seedFacets,/*=NULL*/ 
			     PromTable<int> *lwid_hgid_alias/*=NULL*/ )const
{
  PromFacet      *facet;
  PromGeoNode  *n1;
  gfloat      dot;
  PromList<PromFacet*> facefacets;
  int         jj,count=facets.getCount(),origcount,nnewfaces=0,isseed;
  origcount = count;
  Position posx;
  
  // set types on all facets
  PROM_MAX_FACE_ID = INT_MAX / comm_.np();
  int currID, currID_local = PROM_MAX_FACE_ID * comm_.mype() + 1;
  posx = facets.getHeadPosition(); 
  while( posx && count ) {
    if( seedFacets && !seedFacets->isEmpty() ){
      facet = seedFacets->removeHead();
      if( facet->isMarked() ) continue; // done face
      currID = facet->getID(); assert(currID>0); // imported
      isseed = 1;
    }
    else{
      facet = facets.getNext(posx);
      if( facet->isMarked() ) continue; // done face
      currID = currID_local++;
      assert(currID_local-(PROM_MAX_FACE_ID*comm_.mype())<PROM_MAX_FACE_ID);
      nnewfaces++; isseed = 0;
    }  
    assert(count > 0);
    // new face - this is the seed facet
    PromCoord currnorm(facet->norm_);
    const int N1 = facet->nodes_[3] ? 4 : 3;
    // get least fixed node (leastfixed)
    for ( jj = 1, n1 = dynamic_cast<PromGeoNode*>(facet->nodes_[0]); 
	  jj < N1 ; jj++ ) {
      n1 = !facet->nodes_[jj]->isMoreFixed(n1) ? n1 :
	dynamic_cast<PromGeoNode*>(facet->nodes_[jj]);
    }    
    const PromGeoNode *const leastfixed = n1;
    const int matID = facet->getMaterialID(); assert(facefacets.isEmpty());
    facefacets.addTail(facet);  
    while( !facefacets.isEmpty() && count ) { 
      facet = facefacets.removeHead();
      const int N1 = facet->nodes_[3] ? 4 : 3;
      if( facet->isMarked() ) continue; // dups
      // test for on same 'face'
      dot = currnorm * facet->norm_;
      if( fabs(dot) >= COSTOL && facet->getMaterialID() == matID ) {
	// hack to get Dir. BCs on surfaces
	for ( jj = 1, n1 = dynamic_cast<PromGeoNode*>(facet->nodes_[0]);
	      jj < N1 ; jj++ ) {
	  n1 = !facet->nodes_[jj]->isMoreFixed(n1) ? n1 : 
	    dynamic_cast<PromGeoNode*>(facet->nodes_[jj]);
	} 
	if( n1->isEqualFixed(leastfixed) ) {
	  // got a good one
	  assert( !facet->isMarked() );
	  if( facet->getID() > 0 && facet->getID() != currID && 
	      lwid_hgid_alias && facet->getMaterialID() == matID ){
	    assert(isseed); // ????
	    // alias
	    int ii, tt;
	    if(currID<facet->getID()){ii = currID;tt = facet->getID();}
	    else { tt = currID; ii = facet->getID(); }
	    //aliased facets
	    int add = -lwid_hgid_alias->add( tt, ii );
	    if( add > 0 ){
	      if( add < tt ) lwid_hgid_alias->add( tt, add );
	      else {
		lwid_hgid_alias->add( add, ii ); // undo
		lwid_hgid_alias->add( add, tt ); // chain up 
	      }
	    }
	  }
	  else{
	    facet->setFaceID( currID );// found a facet on this face set type
	  }
	  // add
	  assert(!facet->isMarked());
	  facet->mark(); count--;
	  facet->norm_ = currnorm; // new one, set norm for global tie
	  // for all three (4) nodes of new face - add thier faces to a list
	  for( jj = 0 ; jj < N1 ; jj++ ) {
	    n1 = dynamic_cast<PromGeoNode*>(facet->nodes_[jj]);
	    assert(n1!=NULL);
	    PromList<PromFacet*> *list = &flists[n1->index_()];
	    Position pos2 = list->getHeadPosition();
	    while( pos2 ) {
	      PromFacet *facet2 = list->getNext(pos2); 
	      if( !facet2->isMarked() && 
		  facet->getMaterialID() == facet2->getMaterialID() && 
		  fabs(facet2->norm_*facet->norm_) >= COSTOL ) {
		facefacets.addTail( facet2 ); // add facet to check
	      } 
	    }
	  }
	}
      }
    }     // faces      
    if( options_.verbose_ > 2 ){
      PetscPrintf(PETSC_COMM_SELF,"\t[%d]Done face with %d faux faces(%d/%d)\n",
		  comm_.mype(), currID-(PROM_MAX_FACE_ID*comm_.mype()),
		  count,origcount);
    }
  } // facets
  if(count!=0)PetscPrintf(PETSC_COMM_SELF,"[%d]%s ERROR count=%d\n",
			  comm_.mype(),__FUNCT__);
  
  // collect global faces
  if( isBase() ){ assert(nfaces!=NULL); *nfaces = nnewfaces; }

  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::SendSeeds" 
/* PromGeoGrid::SendSeeds ******************************************
 *
 *  - sends boundary facet types to lower procs
 * 
 *  INPUT: 
 *    - tag:
 *    - lowerProcs:
 *    - facets: 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::SendSeeds( const int tag, PromList<int> &lowerProcs, 
			    PromList<PromFacet*> &facets )const
{
  int        ierr,tt,ii,jj, *pb;
  const int comfact = (prevgrid_==NULL) ? factor_ : prevgrid_->factor_; 
  MPI_Comm comm = (prevgrid_ == NULL) ? gComm_ : prevgrid_->gComm_;
  MPI_Status status; 
  int sbuf[PROM_BSZ]; 

  PromList<MPI_Request*> reqList;
  PromList<int*> buffList;
  PromTable<int> doneIDs; 
  jj = lowerProcs.getCount(); PromFacet *facet;
  while( jj-- ){
    const int proc = lowerProcs.removeHead() - 1; 
    int facide;
    doneIDs.removeAll(); 
    while(1){
      facide = -1;
      pb = sbuf;   
      Position pos = facets.getHeadPosition(); 
      while( pos ) {
	facet = facets.getNext(pos); assert(facet->getID()>0);
	if( doneIDs.findi(facet->getID()) || 
	    (facide != -1 && facide != facet->getID()) ) continue;
	const int N1 = facet->nodes_[3] ? 4 : 3;
	for( ii = 0 ; ii < N1 ; ii++ ) {
	  if( facet->nodes_[ii]->isGhost() && 
	      facet->nodes_[ii]->symbData_->ghostID.proc() == proc ){
	    if( facide == -1 ){
	      facide = facet->getID();
	      facet->norm_.Write( &pb );
	      *pb++ = facet->getMaterialID();
	      *pb++ = facet->getID();
	    }
	    for( tt = 0 ; tt < 3 ; tt++ ) {
	      *pb++ = facet->nodes_[tt]->index_.global();
	    }
	    if(pb-sbuf >= PROM_BSZ) SETERRQ(1,"buffer small");//need to fix!!!
	    break; // sent
	  }
	}
      }
      if( facide != -1 ) {
	doneIDs.add( facide ); 
	MPI_Request *sreq; int *buff;
	ierr = PetscMalloc(sizeof(MPI_Request),&sreq); CHKERRQ(ierr);
	ierr = PetscMalloc((pb-sbuf)*sizeof(int),&buff);CHKERRQ(ierr);
	PetscMemcpy( buff, sbuf, (pb-sbuf) * sizeof(int) );
	// send(tag) < norm, matid, faceid, *<n1, n2, n3> >
	ierr = MPI_Isend(buff, pb-sbuf, MPI_INT, proc/comfact, tag,comm,sreq);
	CHKERRQ(ierr);
	reqList.addTail( sreq );  
	buffList.addTail( buff ); 
      }
      else {
	// send(tag)  sentinal
	ierr = MPI_Send( NULL, 0, MPI_INT, proc/comfact, tag, comm );
	CHKERRQ(ierr);
	break;
      }
    }
  } 
  // wait
  while( !reqList.isEmpty() ) {
    MPI_Request *sreq = reqList.removeHead();
    ierr = MPI_Wait( sreq, &status ); CHKERRQ(ierr);
    ierr = PetscFree( sreq );                  CHKERRQ(ierr);
    int *b = buffList.removeHead();
    ierr = PetscFree( b ); CHKERRQ(ierr);
  }
  
  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "PromGeoGrid::GlobalizeFacets" 
/* PromGeoGrid::GlobalizeFacets ******************************************
 *
 *  - globalize facet aliases
 * 
 *  INPUT: 
 *    - facets:
 *    - lwid_hgid_alias: 
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::GlobalizeFacets( PromList<PromFacet*> &facets, 
				  PromTable<int> &lwid_hgid_alias) const
{
  int        ierr,tt,ii,bsz,old,nnew, *pb; 
  PromTable<int>  lowids;
  Position   pos;
  TablePos   tpos;
  int nCommprocs = comm_.np()/factor_;
  assert(comm_.mype()%factor_==0);
  int sbuf[PROM_BSZ], rbuf[PROM_BSZ]; 
  
  // chain all locals up
  do{
    ii = 0;
    tpos = lwid_hgid_alias.getHeadPosition(); 
    while( tpos ) {
      nnew = lwid_hgid_alias.getNext( tpos, &old ); 
      if( (tt=lwid_hgid_alias.find( nnew )) ) {
	ii++;                assert(nnew != tt); assert(old != tt);
	int add = lwid_hgid_alias.add( tt ,old ); assert(add<0);
      }
    }
  }while(ii);
  // make low list
  tpos = lwid_hgid_alias.getHeadPosition(); 
  while( tpos ) {
    nnew = lwid_hgid_alias.getNext(tpos, &old); assert(nnew != old);
    lowids.add( old ); 
  }
  // alloc bufs, and set < old, new >
  int *local_l_h = NULL, *global_l_h = NULL;
  bsz = lowids.getCount();
  if( bsz ) {
    ierr = PetscMalloc( 2*bsz*sizeof(int),&local_l_h);CHKERRQ(ierr);
    pb = local_l_h;
    tpos = lwid_hgid_alias.getHeadPosition(); 
    while( tpos ) {
      nnew = lwid_hgid_alias.getNext(tpos, &old); assert(nnew != old);
      *pb++ = old; *pb++ = nnew;
    }
    assert(pb-local_l_h == 2*bsz);
  }
  // collect lowids
  ierr = MPI_Allgather(&bsz,1,MPI_INT,rbuf, 1, MPI_INT, gComm_);CHKERRQ(ierr);
  for( ii = tt = sbuf[0] = 0 ; ii < nCommprocs ; ii++ ){ 
    rbuf[ii] *= 2;     tt += rbuf[ii];  // sizes of each
    sbuf[ii+1] = sbuf[ii] + rbuf[ii];   // location to put 
  }
  assert(sbuf[nCommprocs] == tt); const int nh_l = tt/2; 
  if( tt == 0 ) return 0;
  // make new ids
  ierr = PetscMalloc(tt*sizeof(int),&global_l_h);   CHKERRQ(ierr);
  if( options_.verbose_ > 2 ) {
    PetscPrintf(gComm_,"\t[%d]%s Make aliases\n",comm_.mype(),__FUNCT__);
  }
  ierr = MPI_Allgatherv( local_l_h, 2*bsz, MPI_INT, global_l_h, rbuf, sbuf, 
			 MPI_INT, gComm_ ); 
  CHKERRQ(ierr);
  // reset low-->high (make global)
  for( ii = 0 ; ii < 2*nh_l ; ii += 2 ){ 
    old = global_l_h[ii]; nnew = global_l_h[ii+1];
    if( (tt=lwid_hgid_alias.find( old )) && nnew > tt ){
      int add = lwid_hgid_alias.add( nnew, old ); assert(add<0);
    }   
    else lwid_hgid_alias.add( nnew, old );// "old" could be a cur. new
  }
  // chain all locals up
  do{
    ii = 0;
    tpos = lwid_hgid_alias.getHeadPosition(); 
    while( tpos ) {
      nnew = lwid_hgid_alias.getNext( tpos, &old ); 
      if( (tt=lwid_hgid_alias.find( nnew )) ) {
	ii++;                assert(nnew != tt); assert(old != tt);
	int add = lwid_hgid_alias.add( tt ,old ); assert(add<0);
      }
    }
  }while(ii);
  // clean up
  if( local_l_h ) { ierr = PetscFree( local_l_h );  CHKERRQ(ierr); }
  ierr = PetscFree( global_l_h );                   CHKERRQ(ierr);
  
  // reset facet ids
  pos = facets.getHeadPosition(); 
  while( pos ) {
    PromFacet *facet = facets.getNext(pos);
    old = facet->getID();
    nnew = lwid_hgid_alias.find( old ); 
    if( nnew ){assert(nnew >= old); facet->setFaceID(nnew);}
    else nnew =  facet->getID();
    tt = (nnew-1)/PROM_MAX_FACE_ID;
  }
  
  return 0;
} 
 
#define NO_TYPES 0
#define BUFFSZ 45
#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::SetTopoTypes" 
/* PromGeoGrid::SetTopoTypes ******************************************
 *
 *  - set node types
 *
 *  INPUT:
 *    - faces: list of facets
 *    - flists: nodes face lists
 *    - seedFacets: (in/out)
 *    - lwid_hgid_alias: (in/out)
 *
 *  SIDE EFFECTS:
 *    - sets node types, facet colors  and surface colors.
 *    - set nodes "exterior" (needed on fine grid only)
 *
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::SetTopoTypes( PromList<PromFacet*> flists[] )
{ 
  int nplanes, facets[BUFFSZ], faces[BUFFSZ];
  PromCoord norms[BUFFSZ];
  gfloat arr[BUFFSZ][BUFFSZ], dot;
  int kk, jj, found,ierr;
  PromGeoNode *curr;

  // set types
  for( int xx = 0 ; xx < nNodes_ ; xx++ ) {
    curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);

    if( curr->getState() == PROM_NOTUSED ){
      assert(flists==NULL || flists[curr->index_()].isEmpty()); continue;
    }
    if( flists == NULL || flists[curr->index_()].isEmpty() ) {
      // demote 
      switch ( curr->getTopoType() ) {
      case PROM_CORNER : curr->setTopoType(PROM_EDGE); break;
      case PROM_EDGE : curr->setTopoType( PROM_SURFACE ); break;
      case PROM_SURFACE : curr->setTopoType( PROM_INTERIOR ); break;
      default:
	/*no op*/;
      }
    }
    else {  
      // add distict facets of faces (count exterior)
      int nfaces = 0, multmat = FALSE, mat1 = -9;
      while( !flists[curr->index_()].isEmpty() ) {
	PromFacet* facet1 = flists[curr->index_()].removeHead();   
	if( mat1 == -9 ) mat1 = facet1->getMaterialID();
	else if( mat1 != facet1->getMaterialID() ) multmat = TRUE;
	
	for( kk = 0, found = 0 ; kk < nfaces && !found; kk++ ) {
	  if ( facets[kk] == facet1->getID() ) found++;
	}
	if( !found ) {
	  if( nfaces >= BUFFSZ ) SETERRQ(1,"Error num faces > BUFFSZ\n");
	  facets[nfaces] = facet1->getID(); assert( facets[nfaces] > 0 );
	  norms[nfaces] = facet1->norm_;
	  nfaces++;
	}
      }
      assert(nfaces);
      if( multmat || !curr->isExterior() ) curr->setMaterialTopo();
      
      // get number of "planes" passing though node (nplanes)
      for( jj = 0 ; jj < nfaces + 1; jj++ ) faces[jj] = 0;
      for( jj = 0 ; jj < nfaces ; jj++ ) {
	arr[jj][jj] = 1.0;
	for ( kk = jj + 1 ; kk < nfaces ; kk++ ) {
	  dot = norms[kk] * norms[jj];
	  arr[kk][jj] = arr[jj][kk] = dot;
	}
      } 
      for( jj = 0, nplanes = 0 ; jj < nfaces ; jj++ ) {
	if ( faces[jj] ) continue;
	for ( kk = jj ; kk < nfaces ; kk++ ) { 
	  if ( faces[kk] ) continue;
	  if ( fabs(arr[jj][kk]) >= COSTOL ) { 
	    faces[kk] = facets[jj];
	  }
	}
	assert( faces[nplanes] > 0 );
	nplanes++;
      }
      
      // set type
      if( nfaces == 1 ) { // easy 
	assert( !multmat );
	curr->setTopoType( PROM_SURFACE );
	curr->SetFaceIDs( facets[0] ); assert( facets[0] > 0 );
	if( curr->getType() != PROM_SOLID &&
	    curr->symbData_->elemList.getCount() != 0 ){
	  switch( curr->symbData_->adjac.getCount() ) { 
	  case 0: assert(0); 
	  case 1: curr->setTopoType( PROM_CORNER ); break; 
	  case 2: // fallthrough
	  case 3: curr->setTopoType( PROM_EDGE ); break;
	  default: curr->setTopoType( PROM_SURFACE );
	  } 
	}
      }
      else {
	if(nplanes>2 || (nplanes==2 && nfaces>2 && !curr->isMaterialTopo())){
	  curr->setTopoType( PROM_CORNER );
	}
	else {
	  curr->setTopoType( PROM_EDGE );
	  if( nfaces == 0 || nfaces == 1 ) {
	    SETERRQ(1,"One faces on edge.\n");
	  }
	}
	if( nfaces == 2 ) {
	  if( facets[0] > facets[1] ) curr->SetFaceIDs(facets[1], facets[0]);
	  else                        curr->SetFaceIDs(facets[0], facets[1]);
	}
	else {
	  qsort( (void*)facets, nfaces, sizeof(int), prom_intcompar );
	  switch( nfaces) {
	  case 3: 
	    curr->SetFaceIDs( facets[0], facets[1], facets[2]); break;
	  case 4:
	    curr->SetFaceIDs( facets[0],facets[1], facets[2],facets[3]);break;
	  case 5:
	    curr->SetFaceIDs( facets[0], facets[1], facets[2], facets[3],
			      facets[4]);
	    break;
	  default: // 6+ 
	    curr->SetFaceIDs( facets[0], facets[1], facets[2], facets[3],
			      facets[4], facets[5] );
	    if( nfaces > 6 && !curr->isGhost() && options_.verbose_>2 ) 
	      PetscPrintf( PETSC_COMM_SELF,
			   "[%d]High corner facet count(%d) node %d.\n", 
			   comm_.mype(), nfaces, curr->index_(ierr) );
	    break;
	  }
	}
      }
    }
  }
  return 0;
}

#define NUMTYPES 12
#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::MakeNodeList" 
/* PromGeoGrid::MakeNodeList ******************************************
 *
 *  -creates the ordered list from topo types, for MIS
 *
 *  INPUT: 
 *    - list:(out)
 *    - print:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::MakeNodeList(PromList<PromNode*> &list, const bool print)const
{  
  // make lists
  int      ierr,xx,ii,counts[NUMTYPES],gcounts[NUMTYPES],st,end,count;
  PromList<PromNode*> types[NUMTYPES];  
  bool flag, vb = (options_.verbose_>1);
  PromGeoNode *curr;
  Position pos2, pos;

  for( ii = 0 ; ii < NUMTYPES ; ii++ ) counts[ii] = 0;
  ierr = options_.HasOptionName("-geomg_use_geo_hueristics",&flag);
  CHKERRQ(ierr);
  if( flag ) {
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);  
      count = curr->symbData_->adjac.getCount();
      switch( curr->getType() ) {
      case PROM_SOLID: 
	switch( curr->getTopoType() ) {
	case PROM_CORNER       : ii = 0; break;
	case PROM_EDGE         : ii = curr->isMaterialTopo() ? 1 : 2; break;
	case PROM_SURFACE      : ii = curr->isMaterialTopo() ? 3 : 4; break;
	case PROM_INTERIOR     : ii = 11; break;
	default: ii = 11; break;
	}
	break;
      case PROM_PLANE: 
	switch( curr->getTopoType() ) {
	case PROM_CORNER : ii = 5; break; // these are set with "folds"
	case PROM_EDGE : ii = 6; break;
	case PROM_SURFACE : ii = 7; break;
	case PROM_INTERIOR : ii = 11; break; // PROM_INTERIOR
	default: ii = 11; break;
	}
	break;
      case PROM_N_POLY:
	switch( count ) { 
	case 0: // assert(0); flat body
	case 1: // fallthrough
	case 2: // fallthrough
	case 3: ii = 8;  break;
	case 4: // fallthrough
	case 5: // fallthrough
	case 6: ii = 9;  break;
	case 7: // fallthrough
	case 8: // fallthrough
	case 9: // fallthrough
	case 10: ii = 10;  break;
	default: ii = 11;  break; // PROM_INTERIOR
	}
	break;
      default: 
	ii = 11; break;
      }
      // add to a list
      types[ii].addTail( curr );
      if( !curr->isGhost() ) counts[ii]++; 
    }
    // fix for inactive nodes
    if( !isActive() ) {
      for( xx = 0 ; xx < NUMTYPES ; xx++ ) {
	counts[xx] = 0;  types[xx].removeAll();
      }
    }
    // print out -- and kill corners if too many!!!!
    MPI_Allreduce(counts, gcounts, NUMTYPES, MPI_INT, MPI_SUM, gComm_);
    if( (100*gcounts[0]+gcounts[5])/nglobal() > 5 ) {
      flag = FALSE; // simple lists
      if(vb)PetscPrintf(gComm_,
			"\n\t\tWARNING: Killing corners (%d/%d)\n",
			gcounts[0]+gcounts[5],nglobal());
      for( ii = 0 ; ii < NUMTYPES ; ii++ ){
	counts[ii] = 0; types[ii].removeAll();
      }
      // put all in to types[0] (pretend all are corners)
      if( isActive() ) {
	for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
	  curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);  
	  types[0].addTail(curr); counts[0]++;
	}
      }
    }
  }
  else{
    // put all in to types[0] (pretend all are corners)
    if( isActive() ) {
      for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
	curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);  
	types[0].addTail(curr); 
      }
    }
  }

  if( !flag ) { 
    int noint = 0, noext = 0;
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ){ 
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
      if( curr->isExterior() ) noext++;
      else noint++;
    }
    while( !types[0].isEmpty() ) {
      curr = dynamic_cast<PromGeoNode*>(types[0].removeHead());
      curr->setTopoType( PROM_INTERIOR );	
      curr->SetFaceIDs( -9 ); // free for all within class
      curr->setExterior( FALSE );
      count = curr->symbData_->adjac.getCount();
      switch( count ) { 
      case 1: // fallthrough
      case 2: // fallthrough
      case 3: // fallthrough 
      case 4: ii = 2;  break;
      case 5: // fallthrough
      case 6: // fallthrough
      case 7: ii = 3;  break;
      case 8: // fallthrough
      case 9: // fallthrough
      case 10: // fallthrough
      case 11: ii = 4;  break;
      case 12: // fallthrough
      case 13: // fallthrough
      case 14: ii = 6;  break;
      case 15: // fallthrough
      case 16: // fallthrough
      case 17: ii = 7;  break;
      case 18: // fallthrough
      case 19: // fallthrough
      default: ii = 11;  break; // PROM_INTERIOR
      }
      assert(ii>0);
      types[ii].addHead(curr);
      if(!curr->isGhost()) counts[ii]++; 
    }

    while( !types[5].isEmpty() ) {
      curr = dynamic_cast<PromGeoNode*>(types[5].removeHead());
      curr->setTopoType( PROM_INTERIOR );	
      curr->SetFaceIDs( -9 ); // free for all within class
      curr->setExterior( FALSE );
      count = curr->symbData_->adjac.getCount();
      switch( count ) { 
      case 0: // fallthrough
      case 1: // fallthrough
      case 2: // fallthrough
      case 3: ii = 8;  break;
      case 4: // fallthrough
      case 5: ii = 9;  break;
      case 6: // fallthrough
      case 7: // fallthrough
      case 8: ii = 10;  break;
      default: ii = 11;  break; // PROM_INTERIOR
      }      
      types[ii].addHead(curr);
      if(!curr->isGhost()) counts[ii]++; 
    }
    if( isActive() && print && options_.verbose_>1 ) {
      MPI_Allreduce(counts, gcounts, NUMTYPES, MPI_INT, MPI_SUM, gComm_);
    }
  }
  // print
  if( isActive() && print && options_.verbose_>1 ) {
    char typenames[NUMTYPES][16] = {
      "CORNER     ", "EDGE(M)    ", "EDGE       ", "SURFACE(M) ", 
      "SURFACE    ", "C (plane)  ", "E (plane)  ", "S (plane)  ", 
      "C (line)   ", "E (line)   ", "S(line)    ", "INTERIOR   "};
    PetscPrintf(gComm_, "\n\t\t\tMaking Nodes List, %d tot\n",nglobal());
    for( xx = 0 ; xx < NUMTYPES ; xx++ ) {
      if( xx % 2 == 0 ) PetscPrintf(gComm_,"\n");
      PetscPrintf( gComm_, " - type %s:\t%6d nodes\t",
		   typenames[xx], gcounts[xx] );
    } 
    PetscPrintf(gComm_, "\n");
  }

  // make list in 'type' order
  srand(3); // set for determinacy
  ierr = options_.HasOptionName("-prometheus_random_mis", &flag);
  CHKERRQ(ierr);
  if( !flag ) {
    ierr = options_.HasOptionName("-geomg_random_mis", &flag ); CHKERRQ(ierr);
    if( !flag ) {
      ierr = options_.HasOptionName("-aggmg_random_mis", &flag);CHKERRQ(ierr);
    }
  }
  st = 3, end = 17; // topo.ps
  for( ii = 0  ; ii < NUMTYPES ; ii++ ) {
    if( flag ) {
      // randomize, add to lists    
      for( int kk = st, kkk = 2 ; kk < end ; kk++, kkk++ ) {
	pos = types[ii].getHeadPosition();
	while( (pos2=pos) ){
	  curr = dynamic_cast<PromGeoNode*>(types[ii].getNext(pos));
	  // order list - exterior, bound, random
	  bool doit = FALSE;
	  if( NO_TYPES == 0 ){
	    if( kk == end-2 ) {   // last itertion - move fixed up
	      if( curr->isFixed() ) doit = TRUE;// keep exterior in order
	    }
	    else if(curr->getType() == PROM_SOLID && 
		    (curr->isExterior() || curr->isMaterialTopo()))doit = TRUE;
	    else if( rand()%kkk == 0 ) doit = TRUE;
	  }
	  else if( rand()%kkk == 0 ) doit = TRUE;
	  if( doit ) {
	    types[ii].removeAt( pos2 ); 
	    types[ii].addHead( curr ); 
	  }
	}
      }
    }
    else {
      pos = types[ii].getHeadPosition();
      while( (pos2=pos) ){
	curr = dynamic_cast<PromGeoNode*>(types[ii].getNext(pos));
	if( curr->isFixed() || curr->isMaterialTopo() ) {
	  types[ii].removeAt( pos2 ); 
	  types[ii].addHead( curr ); 
	}
      }
    }
    list.addTail( &types[ii] );
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromIsGeoDeleteable" 
/* PromIsGeoDeleteable **********************************
 *
 *  - see if this can delete n1
 *
 *  INPUT:
 *   - curr: 'this'
 *   - n1: node to check
 *   - killit: (out)
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromIsGeoDeleteable( const PromNode * const curr,
			 const PromNode * const n1, bool &deletable )
{
  int ii, jj, num;      
  
  deletable = FALSE;
  if( curr->getNDF() != n1->getNDF() ) return 0; // for now!
  // seperate materials
  if( n1->getTopoType() == PROM_INTERIOR )           deletable = TRUE;
  else if( n1->getTopoType() < curr->getTopoType() )     deletable = FALSE; 
  else { 
    //assert(faceID[0] > 0 || n1->faceID[0] > 0); 
    // edges and and corners
    const PromGeoNode *const gcurr = dynamic_cast<const PromGeoNode*>(curr);
    const int *const cfacid = gcurr->symbGeoData_->faceID;
    const PromGeoNode *const gn1 = dynamic_cast<const PromGeoNode*>(n1);
    const int *const n1facid = gn1->symbGeoData_->faceID;
    num = ii = jj = 0;  
    while( cfacid[ii] != -9 && n1facid[jj] != -9 && ii < 6 && jj < 6 ){
      if( cfacid[ii] == n1facid[jj] ){ num++; ii++; jj++; }
      else if( cfacid[ii] < n1facid[jj] ) ii++;
      else jj++;
    }
    while( n1facid[jj] != -9 && jj < 6 ) jj++;
    while( cfacid[ii] != -9 && ii < 6 ) ii++;
    // test
    if( num <= 1 && ii <= 1 && jj <= 1 ) deletable = TRUE;
    else if( num >= 1 && n1->getTopoType() == PROM_SURFACE ) deletable = TRUE;
    else if( num >= 2 && n1->getTopoType() == PROM_EDGE )  deletable = TRUE;
    // corners ???
  }
  
  return 0;
}

/* PromGeoGrid::Coarsen ***************************************************
 *
 *  - take this mesh and coarsen, to make (almost) max. ind. sets 
 *
 *  INPUT:
 * 
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PET cerror code 
 */
#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::Coarsen" 
int PromGeoGrid::Coarsen()
{
  // insert nodes to the mesh  -skip done nodes
  int              ierr, xx; bool flag; 
  PromGeoNode     *curr; const int tag = getNewTag();
  PromList<PromNode*> nodeList;
  PromGlobalID    *dep2;

  // check files
  char fname[256] = "\0";     
  ierr = options_.HasOptionName("-prometheus_archive_mis",&flag);
  CHKERRQ(ierr); 
  sprintf( fname, "./MIS_lev-%d", getLevel() );
  FILE *file = NULL;  
  if( flag && (file = fopen( fname, "r")) ) {
    fname[0] = '\0'; // will !write
    PetscPrintf(gComm_,"[%d]%s Reading MIS\n",comm_.mype(),__FUNCT__);
    ierr = ReadMISFile( file, xx ); CHKERRQ(ierr); 
    flag = (bool)(xx==1);
    fclose( file );
    file = NULL;
    xx = 0;
  }
  else xx = 1;

  if( xx ) {
    int proc,ii,tt; PromElement::GetWrite_size(tt);
    MPI_Status status;       Position pos;   
    const int reqsz = sizeof(MPI_Request)/sizeof(int),sz=tt+2;// +faces
    PromList<PromGeoNode*> *list; PromList<MPI_Request*> smessages;
    PromTable<PromList<PromGeoNode*>*> procList(57);

    // make node list for coarsening
    ierr = MakeNodeList( nodeList, TRUE );                    CHKERRQ(ierr);

    // globailze topo type and facets
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ){ 
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
      if( curr->isBound()){
	Position pos = curr->symbData_->dependents->getHeadPosition();
	assert(pos!=NULL);
	while( pos ) { 
	  dep2 = curr->symbData_->dependents->getNext( pos ); 
	  proc = dep2->proc();
	  list = procList.find(proc+1);
	  if( list == NULL ) {
	    list = new PromList<PromGeoNode*>(100); CHKNEW(list,1);
	    procList.add( list, proc+1 );
	  }
	  list->addTail( curr );	    
	}
      }
    }
    // isend
    while( !procList.isEmpty() ) {
      list = procList.removeHead(&proc); proc--;
      pos = list->getHeadPosition(); assert(pos!=NULL); tt = 0;
      while( pos ) {
	curr = list->getNext( pos ); // gid
	tt += sz + curr->numFaces();
      }
      MPI_Request *sreq; ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); 
      CHKERRQ(ierr);   	smessages.addTail( sreq );
      int *buff = (int*)(sreq+1), *pb = buff;
      while( !list->isEmpty() ) { 
	curr = list->removeHead();  
	*pb++ = curr->index_.global(); curr->PromElement::Write( &pb ); 
	curr->WriteFaceIDs( pb );
      }
      assert(tt==pb-buff);
      // send(tag) *< gid, ELEM, faces >
      MPI_Isend( buff, pb-buff, MPI_INT, proc/factor_, tag, gComm_, sreq );
      delete list; 
    }
    // recv
    int todo = nNodes_ - nLocalNd_;
    while( todo ) {
      MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
      MPI_Get_count( &status, MPI_INT, &tt ); assert(tt > 0);
      int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
      CHKERRQ(ierr);
      // recv(tag) *< gid, ELEM, faces >
      MPI_Recv( buff, tt, MPI_INT, status.MPI_SOURCE, tag, gComm_, &status);
      while(pb-buff < tt) {
	todo--; assert(todo>=0);
	ii = *pb++;
	curr = dynamic_cast<PromGeoNode*>(ghosts_.find(ii+1));
	assert(curr!=NULL); 
	curr->PromElement::Read( &pb );    curr->ReadFaceIDs( pb );
      }
      assert(tt==pb-buff); PetscFree(buff); 
    }
    // cleanup sends
    while( !smessages.isEmpty() ) {
      MPI_Request *sreq = smessages.removeHead();
      MPI_Wait( sreq, &status );    PetscFree( sreq );
    } 
    MPI_Iprobe(MPI_ANY_SOURCE, tag, gComm_, &ii, &status);assert(!ii);
    // MIS 
    bool mis2; int mis_levels; 
    int *origadjs = NULL;
    ierr = options_.GetOptionInt("-prometheus_mis_levels",&mis_levels, &mis2);
    if( !mis2 ) mis_levels = 1;
    if( mis_levels > 1 && isBase() ) {
      ierr = SquareGraph( &origadjs ); CHKERRQ(ierr);
    }
    else mis2 = FALSE;
    ierr = CreateMIS( nodeList, PromIsGeoDeleteable );CHKERRQ(ierr);
    ierr = UnSquareGraph( &origadjs ); CHKERRQ(ierr);
  }

  // write serial mis
  if( flag && fname[0] != '\0' && comm_.np() == 1 ) {
    PetscPrintf(gComm_,"[%d]%s Writing MIS\n", comm_.mype(), __FUNCT__);
    FILE *file = fopen( fname, "w" ); assert(file!=NULL);
    for( xx = 0 ; xx < nLocalNd_ ; xx++ ){ 
      curr = dynamic_cast<PromGeoNode*>(nodes_[xx]); 
      // write: fid, state
      //ierr=fprintf(file," %d %d",curr->index_.feap(ierr),curr->getState());
      assert(0); // no original IDs!!!!
      CHKERRQ(ierr);
    }
    fprintf( file, "\n" );
    fclose( file );
  }

  return 0;
}

#define PROM_USE_MIS
#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::MakeInitAdjacs"
/* PromGeoGrid::MakeInitAdjacs ******************************
 *
 *  - make approx adjacentcies on coarse nodes. Do not make ghost--ghost.
 *
 *  INPUT:
 *  SIDE EFFECTS:
 *  RETURN:
 *     - PETSc error code
 */
int PromGeoGrid::MakeInitAdjacs()
{
  PromNode             *f1, *f2, *c2, *curr, *fine; 
  int                   xx, nadjs, nnc = 0;
  TablePos              tpos;    

#if defined(PROM_USE_MIS)  
  PromGrid *lgrid = prevgrid_; assert(lgrid!=NULL);
  // set children on last
  for( xx = 0 ; xx < lgrid->nNodes_ ; xx++ ) {
    fine = lgrid->nodes_[xx];
    if( fine->getState() == PROM_DELETED ) {
      f2 = fine->symbData_->parent;  assert(f2!=NULL);
      f2->addChild( fine ); 
    }
  }
#else
  int numreal;
  const int dd = 1, minreal = (nLocalNd_ > dd) ? dd : nLocalNd_-1;
#endif

  // set adjacs for nodes w/o adjacs
  for( xx = nNodes_ - 1, nadjs = 0 ; xx >= 0 ; xx-- ){ // do ghosts first 
    curr = nodes_[xx];  assert(!curr->isDead());  
    fine = curr->symbData_->child; 
    if( !fine ) continue;  // split grid || curr->getState() == PROM_NOTUSED
    
#if defined(PROM_USE_MIS)  
    assert(fine->getState() == PROM_SELECTED);
    for( f2 = dynamic_cast<PromNode*>(fine) ; f2 ; 
	 f2 = dynamic_cast<PromNode*>(f2->getChild()) ) {
      tpos = f2->symbData_->adjac.getHeadPosition();  // look around
      while( tpos ) { 
	f1 = f2->symbData_->adjac.getNext( tpos );
	c2 = f1->getParent(); assert(c2 != NULL && !c2->isDead()); // go up
	if( c2 != curr && curr->symbData_->adjac.add(c2) ){ 
	  c2->symbData_->adjac.add(curr); 
	}
      }
    }
#else  // brute force
    assert( fine->getState() == PROM_SELECTED );
    int targa = fine->symbData_->adjac.getCount(); 
    const int minadj = curr->isGhost() ? targa/2+1 : targa+1;
    if( curr->isGhost() ) {
      numreal = 0; 
      tpos = curr->symbData_->adjac.getHeadPosition();  
      while( tpos ) { 
	f1 = curr->symbData_->adjac.getNext( tpos ); 
	if( !f1->isGhost() ) numreal++; 
      }
    }
    else numreal = 100;
    // look one level
    tpos = fine->symbData_->adjac.getHeadPosition();  
    while( tpos ) { 
      f1 = fine->symbData_->adjac.getNext( tpos );
      // get selected 
      c2 = f1->getParent();              assert(c2 != NULL && !c2->isDead());
      if( c2 != curr && (!c2->isGhost() || !curr->isGhost()) ) {
	if( curr->symbData_->adjac.add(c2) ){ 
	  c2->symbData_->adjac.add(curr); numreal++;
	}
      }
    }
    // look two level  
    if( numreal < minreal || curr->symbData_->adjac.getCount() < minadj ) {
      PromTable<PromNode*>  finelist(91);
      TablePos tpos2 = fine->symbData_->adjac.getHeadPosition(); 
      assert(tpos2!=0);
      while( tpos2 ) {
	f1 = fine->symbData_->adjac.getNext(tpos2);
	TablePos tpos = f1->symbData_->adjac.getHeadPosition(); 
	assert(tpos!=0);
	while( tpos ) {
	  f2 = f1->symbData_->adjac.getNext(tpos); 
	  if( fine->symbData_->adjac.find(f2) || f2 == fine ) continue;
	  finelist.add( f2 );  
	  // get selected 
	  c2 = f2->getParent();  assert(c2!=NULL && !c2->isDead());
	  if( c2 != curr && !(c2->isGhost() && curr->isGhost()) ) {
	    if( curr->symbData_->adjac.add(c2) ){ 
	      c2->symbData_->adjac.add(curr); numreal++;
	    }
	  }
	}
      }
      // look three level
      if( !finelist.isEmpty() 
	  && (numreal<minreal || curr->symbData_->adjac.getCount()<minadj)){
	if( !curr->isGhost() && options_.verbose_ > 2 ) 
	  PetscPrintf(PETSC_COMM_WORLD,
		      "[%d]WARNING:%s %d) Deep search for adj. (%d < %d)\n",
		      comm_.mype(),__FUNCT__,xx,
		      curr->symbData_->adjac.getCount(),minadj);
	TablePos tpos2 = finelist.getHeadPosition(); assert(tpos2!=0);
	while(tpos2 /* && numreal < minreal&&curr->symbData_->adjac.getCount() < minadj */ ){
	  f1 = finelist.getNext(tpos2);
	  TablePos tpos = f1->symbData_->adjac.getHeadPosition(); //assert(tpos!=0);
	  while(tpos /* && numreal < minreal&&curr->symbData_->adjac.getCount() < minadj */ ){
	    f2 = f1->symbData_->adjac.getNext(tpos);
	    if( f2 == fine ) continue;
	    if(fine->symbData_->adjac.find(f2) || finelist.find(f2))continue;
	    assert( f2 != fine );
	    if( f2->getState() == PROM_SELECTED ) {
	      c2 = dynamic_cast<PromNode*>(f2->symbData_->parent);
	      assert(c2 != NULL && !c2->isDead());
	      if( c2 != curr && !(c2->isGhost() && curr->isGhost()) ) {
		if( curr->symbData_->adjac.add(c2) ){
		  c2->symbData_->adjac.add(curr); numreal++; 
		}
	      }
	    }
	  }
	}
      }
    }
#endif
    if( curr->symbData_->adjac.isEmpty() ) {
      if( !curr->isGhost() ) {
	PetscPrintf(PETSC_COMM_SELF,"[%d]WARNING:%s()%s %d/%d w/o adj,(%d)\n",
		    comm_.mype(),__FUNCT__,curr->isGhost()?"Ghost":"**Non-ghost**",
		    xx, nLocalNd_, fine->symbData_->adjac.getCount() );
	// put in arbitrary edge -- should not happen ever
	for( int yy = 0 ; yy < nLocalNd_ ; yy++ ){ 
	  PromNode *n1 = nodes_[yy];  
	  if( curr != n1 ){
	    curr->symbData_->adjac.add(n1); n1->symbData_->adjac.add(curr);
	    break;
	  }
	}
      }
    }
    nadjs += curr->symbData_->adjac.getCount(); nnc++;
  }
  if(options_.verbose_>2)PetscPrintf(PETSC_COMM_SELF,"\t\t\t[%d]:%s av adj = %g\n",
				      comm_.mype(),__FUNCT__,(float)nadjs/(float)nnc);
  
#if defined(PROM_USE_MIS)  
  lgrid = prevgrid_; assert(lgrid!=NULL);
  // unset children on last
  for( xx = 0 ; xx < lgrid->nNodes_ ; xx++ ) {
    fine = lgrid->nodes_[xx]; fine->ClearChildren();
  }
#endif

  return 0;
}

#undef __FUNCT__    
#define __FUNCT__ "PromGeoGrid::CreateGeometricCoarseSpace" 
/* PromGeoGrid::CreateGeometricCoarseSpace ****************************
 *
 *  - take this mesh and create coarse grid space 
 *    + triangulate (coarse)
 *    + find node elements
 *    + fix mesh (coarse)
 *
 *  INPUT:
 *    - lastGrid: fine grid
 *
 *  SIDE EFFECTS:
 *    - make "parents"
 *
 *  RETURN:
 *    - PETSc error code 
 */ 
int PromGeoGrid::CreateGeometricCoarseSpace( PromGeoGrid * const lastGrid )
{
  // insert nodes to the mesh  -skip done nodes
  int lost, ierr;

  PromCoord minc, maxc, cfactor( 11.0, 11.5, 12.0 ); 
  MPI_Comm comm = (options_.verbose_>2) ? PETSC_COMM_SELF : lastGrid->gComm_;
  const int lev2 = getLevel(), lev1 = lev2 - 1;

  // offset disconnected parts (contact problems can have dup. nodes)
  ierr = OffsetParts( lastGrid, cfactor, minc, maxc, FALSE );  CHKERRQ(ierr);
  
  // triangulate coarser grid -- can inactive (does redundent work), no mpi
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_TRIANGULATE]);
  assert(elemList_.isEmpty());
  ierr = Triangulate();                   CHKERRQ(ierr);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_TRIANGULATE]);

  if(options_.verbose_>1)PetscPrintf(comm,
				      "[%d]Triangulate(%d)--done\n[%d]findNodeElem(%d).",
				      comm_.mype(), lev2, comm_.mype(), lev1 );
  
  // set element pointer in nodes of next grid -- no mpi
  ierr = lastGrid->findNodeElems( lost );      CHKERRQ(ierr);
  assert( lost == 0 ); // findNodeElems gets all
  if(options_.verbose_>1)PetscPrintf(comm,
				      "[%d]findNodeElem(%d)--done\n[%d]FixMesh(%d)...",
				      comm_.mype(), lev1, comm_.mype(), lev2);

  // fix mesh on coarser grid -- no mpi
  perf_mon_.EventBegin(PromContext::prom_logs_[LOG_FIXMESH]); 
  ierr = FixMesh( &elemList_  );                            CHKERRQ(ierr);
  perf_mon_.EventEnd(PromContext::prom_logs_[LOG_FIXMESH]);
  if(options_.verbose_>1)PetscPrintf(comm,
				      "[%d]FixMesh(%d) -- done\n", 
				      comm_.mype(), lev2);
  ierr = CleanElemList(); CHKERRQ(ierr);
  // put grid back
  ierr = OffsetParts(lastGrid, cfactor*-1.0, minc, maxc, TRUE ); CHKERRQ(ierr);
  // remove all traces of elements on LAST gri
  ierr = lastGrid->DeleteElemList(); CHKERRQ(ierr);
  // output THIS grid
  if( options_.files_ && isActive() ){ 
    assert( !elemList_.isEmpty() );
    ierr = WriteFEAP( 10000*getLevel()+comm_.mype(), gComm_, &elemList_ );
    CHKERRQ(ierr);  
  }

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::DeleteElemList"
/* PromGeoGrid::DeleteElemList ******************************
 *
 */
int PromGeoGrid::DeleteElemList()
{
  //
  while( !elemList_.isEmpty() ) {
    PromPoly  *elem = elemList_.removeHead();
    if( !elem->isDead() ){    
      // delete elems base grids elems are not in the bank.
      if( isBase() ) delete elem; 
      else Kill( dynamic_cast<PromDTet*>(elem) );
    } 
  }
  // remove pointerst to these elements
  for ( int xx = 0 ; xx < nNodes_ ; xx++ ){
    PromGeoNode *curr = dynamic_cast<PromGeoNode*>(nodes_[xx]);
    if( curr->symbGeoData_ != NULL ) {
      curr->symbGeoData_->parElem = NULL; assert(curr->symbData_ != NULL);
      curr->symbData_->elemList.removeAll();
    }
  }
  
  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::PrepForSolve"
/* PromGeoGrid::PrepForSolve ***************************************
 *
 */ 
int PromGeoGrid::PrepForSolve()
{
  int ierr;

  // clear geo symb data
  ierr = DeleteElemList();  CHKERRQ(ierr);     assert(elemList_.isEmpty());
  if( !isTop() && isActive() ){
    ierr = SetParents();  CHKERRQ(ierr);
  }

  if( !isTop() && isActive() ) {
    PromGeoGrid *ng = dynamic_cast<PromGeoGrid*>(nextgrid_);
    // make exact(numeric)
    if(options_.verbose_>1)PetscPrintf( gComm_, "[%d]%s MakeExactAdjac(%d)\n",
					comm_.mype(),__FUNCT__,ng->getLevel());
    perf_mon_.EventBegin(PromContext::prom_logs_[LOG_EXACTADJ]);

    ierr = ng->MakeExactAdjac( this );       CHKERRQ(ierr);
    perf_mon_.EventEnd(PromContext::prom_logs_[LOG_EXACTADJ]);
  }

  for(int i=0;i<PROM_NUMENV;i++)if(env_[i]!=NULL){Kill(env_[i]); env_[i]=NULL;}
  ierr = PromGrid::PrepForSolve();  CHKERRQ(ierr);
  ierr = DeleteTets();      CHKERRQ(ierr);
  ierr = DeleteFacets();    CHKERRQ(ierr);

  return 0;
}

#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::SetParents"  
/* PromGeoGrid::SetParents *****************************************
 *
 *  - make node parents w/ elements on coarse grid
 *
 *  INPUT:
 *
 *  SIDE EFFECTS:
 *  RETURN:
 *    - PETSc error code
 */
int PromGeoGrid::SetParents()
{ 
  int       xx,ierr,locid,todo,proc,ii=sizeof(int); xx = sizeof(shpfloat);
  int       tt=sizeof(MPI_Request);
  const int sisz = xx/ii, reqsz = (tt/xx + ((tt%xx == 0) ? 0 : 1)) * sisz;
  PromNode *curr;  const int tag = getNewTag();
  MPI_Status   status;
  Position     pos;
  PromList<PromNode*> *list;
  PromList<MPI_Request*> smessages; 
  PromTable<PromList<PromNode*>*> proc_list(57);
  PromGlobalID *dep;

  // boundary node - send dep updates, move local procs to master proc 
  for( xx = 0 ; xx < nLocalNd_ ; xx++ ){
    curr = nodes_[xx]; assert( curr->getState()!=PROM_NOTUSED);
    // get array of parent data and send
    if( curr->isBound() ) {
      pos = curr->symbData_->dependents->getHeadPosition(); assert(pos!=NULL);
      while( pos ) {
	dep = curr->symbData_->dependents->getNext( pos );
	proc = dep->proc();
	if( (list=proc_list.find(proc+1)) == NULL ) {
	  list = new PromList<PromNode*>(100); CHKNEW(list,1);
	  proc_list.add( list, proc+1 );
	}
	list->addTail( curr ); list->addTail( (PromNode*)dep ); // hack
      }
    }
  }
  // send data
  while( !proc_list.isEmpty() ) {
    list = proc_list.removeHead( &proc ); proc--;
    pos = list->getHeadPosition(); assert(pos!=NULL); tt = 1;
    while( pos ) {
      curr = list->getNext( pos ); 
      dep = (PromGlobalID*)list->getNext( pos ); // hack 
      tt++; // dep->locid
      if(tt%sisz != 0) tt++;  
      curr->GetWrite_size(ii); tt += ii; 
    }
    MPI_Request *sreq;
    ierr = PetscMalloc((tt+reqsz)*sizeof(int),&sreq); CHKERRQ(ierr);
    smessages.addTail( sreq );
    int *buff = (int*)sreq + reqsz, *pb = buff;
    assert((size_t)pb%sizeof(shpfloat) == 0);
    *pb++ = list->getCount()/2;
    while( !list->isEmpty() ) {
      curr = list->removeHead(); 
      dep = (PromGlobalID*)list->removeHead( ); // hack
      *pb++ = dep->local();
      if((size_t)pb%sizeof(shpfloat) != 0) *pb++ = -1;
      curr->Write( &pb ); 
    }
    assert(tt == pb-buff);
    // send(tag) < n, n*< lid, [-1,] node > >
    MPI_Isend( buff, tt, MPI_INT, proc/factor_, tag, gComm_, sreq );
    delete list;
    assert(proc%factor_ == 0);
  }

  // recieve gid, ... for ghosts update
  todo = nNodes_ - nLocalNd_;
  while( todo ) {
    MPI_Probe( MPI_ANY_SOURCE, tag, gComm_, &status );
    MPI_Get_count( &status, MPI_INT, &tt ); proc = status.MPI_SOURCE*factor_;
    int *buff; ierr = PetscMalloc(tt*sizeof(int),&buff); int *pb = buff;
    CHKERRQ(ierr);
    // receive(tag) < n, n*< lid, [-1,] node > >
    MPI_Recv( buff, tt, MPI_INT, MPI_ANY_SOURCE, tag, gComm_, &status);
    int nn = *pb++;   todo -= nn; assert(todo>=0);
    while( nn-- ) {
      locid = *pb++;  assert(locid>=nLocalNd_&&locid<nNodes_); // ghost
      if((size_t)pb%sizeof(shpfloat) != 0) pb++; // make even
      // get ncurr and set
      curr = nodes_[locid];
      ierr = curr->Read( parBank_, &pb );  CHKERRQ(ierr);
      curr->index_ = locid;      // reset local
    }
    assert(pb-buff == tt);   PetscFree(buff); 
  }
  // free
  while( !smessages.isEmpty() ) {
    MPI_Request *sreq = smessages.removeHead();
    MPI_Wait( sreq, &status );    PetscFree( sreq );
  }

  return 0;
}


#undef __FUNCT__  
#define __FUNCT__ "PromGeoGrid::CleanElemList" 
/* PromGeoGrid::CleanElemList *********************************************
 *
 *  - cleans 'elemList' of dead elems and adds them onto free list 
 *
 *  - returns number killed elements
 */
int PromGeoGrid::CleanElemList()
{   
  assert(this!=NULL); assert( !isBase() );
  int count = 0;

  Position pos2, pos = elemList_.getHeadPosition();
  while ( (pos2 = pos) != NULL ) { 
    PromDTet *elem = dynamic_cast<PromDTet*>(elemList_.getNext(pos));
    //  remove dead elems
    if ( elem->isDead() ) { elemList_.removeAt( pos2 );  count++; }
  }

  return count;
}
